/* 
 *  Author: Chris Lord <chris@openedhand.com>
 *
 *  Copyright (c) 2007 OpenedHand Ltd - http://www.openedhand.com/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */

#include "jana-ecal-component.h"
#include "jana-ecal-event.h"
#include "jana-ecal-time.h"
#include <libjana/jana-time.h>
#include <libjana/jana-component.h>
#include <libjana/jana-event.h>
#include <libjana/jana-utils.h>

static void event_interface_init (gpointer g_iface, gpointer iface_data);
static void component_interface_init (gpointer g_iface, gpointer iface_data);

static JanaComponentType	component_get_component_type
							(JanaComponent *self);
static gboolean			component_is_fully_represented
							(JanaComponent *self);

static gchar * 	event_get_summary	(JanaEvent *self);
static gchar * 	event_get_description	(JanaEvent *self);
static gchar * 	event_get_location	(JanaEvent *self);
static JanaTime * 	event_get_start	(JanaEvent *self);
static JanaTime *	event_get_end	(JanaEvent *self);
static gchar **	event_get_categories	(JanaEvent *self);

static gboolean	event_supports_alarm	(JanaEvent *self);
static gboolean	event_has_alarm		(JanaEvent *self);
static JanaTime *event_get_alarm_time	(JanaEvent *self);

static gboolean		event_supports_recurrence	(JanaEvent *self);
static gboolean		event_has_recurrence		(JanaEvent *self);
static JanaRecurrence *	event_get_recurrence		(JanaEvent *self);

static gboolean	event_supports_exceptions	(JanaEvent *self);
static gboolean	event_has_exceptions	(JanaEvent *self);
static GList *	event_get_exceptions	(JanaEvent *self);

static void	event_set_summary	(JanaEvent *self,
					 const gchar *summary);
static void	event_set_description	(JanaEvent *self,
					 const gchar *description);
static void	event_set_location	(JanaEvent *self,
					 const gchar *location);
static void	event_set_start		(JanaEvent *self, JanaTime *start);
static void	event_set_end		(JanaEvent *self, JanaTime *end);
static void	event_set_categories	(JanaEvent *self,
					 const gchar **categories);

static void	event_set_alarm		(JanaEvent *self, JanaTime *time);

static void	event_set_recurrence	(JanaEvent *self,
					 const JanaRecurrence *recurrence);

static void	event_set_exceptions	(JanaEvent *self, GList *exceptions);

G_DEFINE_TYPE_WITH_CODE (JanaEcalEvent, 
			 jana_ecal_event, 
			 JANA_ECAL_TYPE_COMPONENT,
			 G_IMPLEMENT_INTERFACE (JANA_TYPE_COMPONENT,
						component_interface_init)
			 G_IMPLEMENT_INTERFACE (JANA_TYPE_EVENT,
						event_interface_init));

#define EVENT_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
			  JANA_ECAL_TYPE_EVENT, JanaEcalEventPrivate))


typedef struct _JanaEcalEventPrivate JanaEcalEventPrivate;

struct _JanaEcalEventPrivate
{
};

static void
jana_ecal_event_get_property (GObject *object, guint property_id,
			      GValue *value, GParamSpec *pspec)
{
	/*JanaEcalEventPrivate *priv = EVENT_PRIVATE (object);*/

	switch (property_id) {
	    default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
jana_ecal_event_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
	/*JanaEcalEventPrivate *priv = EVENT_PRIVATE (object);*/

	switch (property_id) {
	    default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
jana_ecal_event_dispose (GObject *object)
{
	if (G_OBJECT_CLASS (jana_ecal_event_parent_class)->dispose)
		G_OBJECT_CLASS (jana_ecal_event_parent_class)->dispose (object);
}

static void
jana_ecal_event_finalize (GObject *object)
{
	G_OBJECT_CLASS (jana_ecal_event_parent_class)->finalize (object);
}

static void
component_interface_init (gpointer g_iface, gpointer iface_data)
{
	JanaComponentInterface *iface = (JanaComponentInterface *)g_iface;
	
	iface->get_component_type = component_get_component_type;
	iface->is_fully_represented = component_is_fully_represented;
}

static void
event_interface_init (gpointer g_iface, gpointer iface_data)
{
	JanaEventInterface *iface = (JanaEventInterface *)g_iface;
	
	iface->get_summary = event_get_summary;
	iface->get_description = event_get_description;
	iface->get_location = event_get_location;
	iface->get_start = event_get_start;
	iface->get_end = event_get_end;
	iface->get_categories = event_get_categories;
	
	iface->supports_alarm = event_supports_alarm;
	iface->has_alarm = event_has_alarm;
	iface->get_alarm_time = event_get_alarm_time;
	
	iface->supports_recurrence = event_supports_recurrence;
	iface->has_recurrence = event_has_recurrence;
	iface->get_recurrence = event_get_recurrence;
	
	iface->supports_exceptions = event_supports_exceptions;
	iface->has_exceptions = event_has_exceptions;
	iface->get_exceptions = event_get_exceptions;
	
	iface->set_summary = event_set_summary;
	iface->set_description = event_set_description;
	iface->set_location = event_set_location;
	iface->set_start = event_set_start;
	iface->set_end = event_set_end;
	iface->set_categories = event_set_categories;
	iface->set_alarm = event_set_alarm;
	iface->set_recurrence = event_set_recurrence;
	iface->set_exceptions = event_set_exceptions;
}

static void
jana_ecal_event_class_init (JanaEcalEventClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	/* Don't add a zero-size private, causes memory corruption issues */
	/*g_type_class_add_private (klass, sizeof (JanaEcalEventPrivate));*/

	object_class->get_property = jana_ecal_event_get_property;
	object_class->set_property = jana_ecal_event_set_property;
	object_class->dispose = jana_ecal_event_dispose;
	object_class->finalize = jana_ecal_event_finalize;
}

static void
jana_ecal_event_init (JanaEcalEvent *self)
{
}

JanaEvent *
jana_ecal_event_new ()
{
	ECalComponent *comp = e_cal_component_new ();
	e_cal_component_set_icalcomponent (comp,
		icalcomponent_new (ICAL_VEVENT_COMPONENT));

	return JANA_EVENT (g_object_new (JANA_ECAL_TYPE_EVENT,
		"ecalcomp", comp, NULL));
}

JanaEvent *
jana_ecal_event_new_from_ecalcomp (ECalComponent *event)
{
	return JANA_EVENT (g_object_new (JANA_ECAL_TYPE_EVENT,
		"ecalcomp", event, NULL));
}

static JanaComponentType
component_get_component_type (JanaComponent *self)
{
	return JANA_COMPONENT_EVENT;
}

static gboolean
component_is_fully_represented (JanaComponent *self)
{
	gboolean result = TRUE;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);

	/* TODO: Implement recurrences */
	if (e_cal_component_has_rrules (comp))
		result = FALSE;
	
	if (e_cal_component_has_rdates (comp))
		result = FALSE;

	/* TODO: Implement exception dates */
	if (e_cal_component_has_exdates (comp))
		result = FALSE;

	if (e_cal_component_has_exrules (comp))
		result = FALSE;
	
	if (e_cal_component_has_organizer (comp))
		result = FALSE;
	
	if (e_cal_component_has_attendees (comp))
		result = FALSE;
	
	if (e_cal_component_has_attachments (comp))
		result = FALSE;
	
	/* TODO: Implement alarms */
	if (e_cal_component_has_alarms (comp))
		result = FALSE;

	/* TODO: There are probably more parts of ECalComponent/icalcomponent
	 *       that aren't supported. Need to either add more support at the
	 *       JanaEvent level, or do more in-depth checking here. This will
	 *       catch the most important aspects at least.
	 */
	
	g_object_unref (comp);

	return result;
}

static gchar *
event_get_summary (JanaEvent *self)
{
	ECalComponentText summary;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	e_cal_component_get_summary (comp, &summary);
	g_object_unref (comp);
	
	if (summary.value) return g_strdup (summary.value);
	/*else if (summary.altrep) return g_strdup (summary.altrep);*/
	else return NULL;
}

static gchar *
event_get_description (JanaEvent *self)
{
	GSList *desc_list;
	ECalComponentText *desc;
	ECalComponent *comp;
	gchar *desc_copy = NULL;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	/* Note, that only journal components are allowed to have more than one
	 * description. It's ok to just take the first here.
	 */
	e_cal_component_get_description_list (comp, &desc_list);
	if (desc_list) {
		desc = desc_list->data;
		if (desc->value)
			desc_copy = g_strdup (desc->value);
	}
	
	e_cal_component_free_text_list (desc_list);
	
	g_object_unref (comp);
	
	return desc_copy;
}

static gchar *
event_get_location (JanaEvent *self)
{
	const char *location;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	e_cal_component_get_location (comp, &location);
	
	g_object_unref (comp);
	
	return location ? g_strdup (location) : NULL;
}

static JanaTime *
event_get_start (JanaEvent *self)
{
	ECalComponentDateTime dt;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);

	dt.value = NULL;	
	e_cal_component_get_dtstart (comp, &dt);
	g_object_unref (comp);
	
	if (dt.value) {
		JanaTime *start = jana_ecal_time_new_from_icaltime (dt.value);
		e_cal_component_free_datetime (&dt);
		return start;
	} else
		return NULL;
}

static JanaTime *
event_get_end (JanaEvent *self)
{
	ECalComponentDateTime dt;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	dt.value = NULL;
	e_cal_component_get_dtend (comp, &dt);
	g_object_unref (comp);
	
	if (dt.value) {
		JanaTime *end = jana_ecal_time_new_from_icaltime (dt.value);
		e_cal_component_free_datetime (&dt);
		return end;
	} else
		return NULL;
}

static gchar **
event_get_categories (JanaEvent *self)
{
	const char *categories;
	gchar **category_list;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	e_cal_component_get_categories (comp, &categories);
	g_object_unref (comp);

	if (categories) {
		category_list = g_strsplit (categories, ",", 0);
		return category_list;
	} else
		return NULL;
}

static gboolean
event_supports_alarm (JanaEvent *self)
{
	/* TODO: Implement alarms */
	return FALSE;
}

static gboolean
event_has_alarm (JanaEvent *self)
{
	/* TODO: Implement alarms */
	return FALSE;
}

static JanaTime *
event_get_alarm_time (JanaEvent *self)
{
	/* TODO: Implement alarms */
	return NULL;
}

static gboolean
event_supports_recurrence (JanaEvent *self)
{
	/* TODO: Implement recurrence */
	return FALSE;
}

static gboolean
event_has_recurrence (JanaEvent *self)
{
	/* TODO: Implement recurrence */
	return FALSE;
}

static JanaRecurrence *
event_get_recurrence (JanaEvent *self)
{
	/* TODO: Implement recurrence */
	return NULL;
}

static gboolean
event_supports_exceptions (JanaEvent *self)
{
	/* TODO: Implement exceptions */
	return FALSE;
}

static gboolean
event_has_exceptions (JanaEvent *self)
{
	/* TODO: Implement exceptions */
	return FALSE;
}

static GList *
event_get_exceptions (JanaEvent *self)
{
	/* TODO: Implement exceptions */
	return NULL;
}

static void
event_set_summary (JanaEvent *self, const gchar *summary)
{
	ECalComponentText summary_text;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	summary_text.value = summary;
	summary_text.altrep = NULL;
	
	e_cal_component_set_summary (comp, &summary_text);
	g_object_unref (comp);
}

static void
event_set_description (JanaEvent *self, const gchar *description)
{
	ECalComponentText desc_text;
	GSList *text_list;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	desc_text.value = description;
	desc_text.altrep = NULL;
	text_list = g_slist_append (NULL, &desc_text);
	
	e_cal_component_set_description_list (comp, text_list);
	
	g_object_unref (comp);
	
	g_slist_free (text_list);
}

static void
event_set_location (JanaEvent *self, const gchar *location)
{
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	e_cal_component_set_location (comp, location);
	g_object_unref (comp);
}

static void
event_set_start (JanaEvent *self, JanaTime *start)
{
	JanaEcalTime *time;
	icaltimetype *icaltime;
	ECalComponentDateTime dt;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	if (JANA_ECAL_IS_TIME (start)) {
		time = g_object_ref (JANA_ECAL_TIME (start));
	} else {
		time = JANA_ECAL_TIME (jana_utils_time_copy (
			start, jana_ecal_time_new ()));
	}
	
	g_object_get (time, "icaltime", &icaltime, NULL);
	
	/* vaue, tzid */
	dt.value = icaltime;
	if (icaltime_is_utc (*icaltime))
		dt.tzid = "UTC";
	else
		dt.tzid = (const char *)icaltimezone_get_tzid (
			(icaltimezone *)icaltime->zone);
	
	e_cal_component_set_dtstart (comp, &dt);
	g_object_unref (comp);
	
	g_object_unref (time);
}

static void
event_set_end (JanaEvent *self, JanaTime *end)
{
	JanaEcalTime *time;
	icaltimetype *icaltime;
	ECalComponentDateTime dt;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	if (JANA_ECAL_IS_TIME (end)) {
		time = g_object_ref (JANA_ECAL_TIME (end));
	} else {
		time = JANA_ECAL_TIME (jana_utils_time_copy (
			end, jana_ecal_time_new ()));
	}
	
	g_object_get (time, "icaltime", &icaltime, NULL);
	
	/* vaue, tzid */
	dt.value = icaltime;
	if (icaltime_is_utc (*icaltime))
		dt.tzid = "UTC";
	else
		dt.tzid = (const char *)icaltimezone_get_tzid (
			(icaltimezone *)icaltime->zone);
	
	e_cal_component_set_dtend (comp, &dt);
	g_object_unref (comp);
	
	g_object_unref (time);
}

static void
event_set_categories (JanaEvent *self, const gchar **categories)
{
	gchar *categories_joined;
	ECalComponent *comp;
	g_object_get (self, "ecalcomp", &comp, NULL);
	
	categories_joined = g_strjoinv (",", (gchar **)categories);
	/* Note: Setting an empty categories list on an event without a
	 *       category set will cause libical to crash.
	 */
	if ((!categories_joined) || (categories_joined[0] == '\0')) {
		const gchar *categories;
		e_cal_component_get_categories (comp, &categories);
		if (categories)
			e_cal_component_set_categories (comp, NULL);
	} else {
		e_cal_component_set_categories (comp, categories_joined);
	}
	g_free (categories_joined);
	g_object_unref (comp);
}

static void
event_set_alarm (JanaEvent *self, JanaTime *time)
{
	/* TODO: Implement alarms */
}

static void
event_set_recurrence (JanaEvent *self, const JanaRecurrence *recurrence)
{
	/* TODO: Implement recurrence */
}

static void
event_set_exceptions (JanaEvent *self, GList *exceptions)
{
	/* TODO: Implement exceptions */
}

