/* 
 *  Author: Chris Lord <chris@openedhand.com>
 *
 *  Copyright (c) 2007 OpenedHand Ltd - http://www.openedhand.com/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */
 
#include <time.h>
#include "jana-utils.h"

static const guint8 days_in_month[] =
	{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

gboolean
jana_utils_time_is_leap_year (guint16 year)
{
	/* See Wikipedia 'leap years':
	 *	if year modulo 400 is 0 then leap
	 *	 else if year modulo 100 is 0 then no_leap
	 *	 else if year modulo 4 is 0 then leap
	 *	 else no_leap
 	 */
 	if ((!(year % 400)) || ((year % 100) && (!(year % 4))))
 		return TRUE;
 	else
 		return FALSE;
}

guint8
jana_utils_time_days_in_month (guint16 year, guint8 month)
{
	if ((month == 2) && (jana_utils_time_is_leap_year (year)))
		return days_in_month[month-1] + 1;
	else
		return days_in_month[month-1];
}

GDateWeekday
jana_utils_time_day_of_week (JanaTime *time)
{
	GDateWeekday day;
	GDate *date;
	
	/* TODO: Write function to do this directly on a JanaTime */
	date = jana_utils_time_to_gdate (time);
	day = g_date_get_weekday (date);
	g_date_free (date);
	
	return day;
}

guint
jana_utils_time_week_of_year (JanaTime *time, gboolean week_starts_monday)
{
	guint week;
	GDate *date;
	
	/* TODO: Write function to do this directly on a JanaTime */
	date = jana_utils_time_to_gdate (time);
	if (week_starts_monday)
		week = g_date_get_monday_week_of_year (date);
	else
		week = g_date_get_sunday_week_of_year (date);
	g_date_free (date);
	
	return week;
}

GDateWeekday
jana_utils_time_set_start_of_week (JanaTime *start)
{
	GDateWeekday day = jana_utils_time_day_of_week (start);
	if (day)
		jana_time_set_day (start,
			jana_time_get_day (start) - (day - 1));
	
	return day;
}

GDateWeekday
jana_utils_time_set_end_of_week (JanaTime *end)
{
	GDateWeekday day = jana_utils_time_day_of_week (end);
	if (day)
		jana_time_set_day (end,
			jana_time_get_day (end) + (7 - day));

	return day;
}

GDate *
jana_utils_time_to_gdate (JanaTime *time)
{
	GDate *date = g_date_new ();
	g_date_set_dmy (date, jana_time_get_day (time),
		jana_time_get_month (time),
		jana_time_get_year (time));

	return date;
}

gint
jana_utils_time_compare (JanaTime *time1, JanaTime *time2, gboolean date_only)
{
	JanaTime *time2_corrected;
	gint time1_part;
	guint8 seconds, minutes, hours, day, month;
	guint16 year;
	
	/* Get time2 relative to time1 */
	if (jana_time_get_offset (time1) == jana_time_get_offset (time2))
		time2_corrected = g_object_ref (time2);
	else {
		time2_corrected = jana_time_duplicate (time2);
		jana_time_set_offset (time2_corrected,
			jana_time_get_offset (time1));
	}
	year = jana_time_get_year (time2_corrected);
	month = jana_time_get_month (time2_corrected);
	day = jana_time_get_day (time2_corrected);
	if (!jana_time_get_isdate (time2_corrected)) {
		hours = jana_time_get_hours (time2_corrected);
		minutes = jana_time_get_minutes (time2_corrected);
		seconds = jana_time_get_seconds (time2_corrected);
	} else {
		hours = 0; minutes = 0; seconds = 0;
	}
	g_object_unref (time2_corrected);
	
	/* Check years */
	time1_part = jana_time_get_year (time1);
	if (time1_part < year)
		return -1;
	else if (time1_part > year)
		return 1;
	
	/* Check months */
	time1_part = jana_time_get_month (time1);
	if (time1_part < month)
		return -1;
	else if (time1_part > month)
		return 1;
	
	/* Check days */
	time1_part = jana_time_get_day (time1);
	if (time1_part < day)
		return -1;
	else if (time1_part > day)
		return 1;
	
	if (date_only) return 0;
	if (jana_time_get_isdate (time1)) {
		if (hours || minutes || seconds)
			return -1;
		else
			return 0;
	}
	
	/* Check hours */
	time1_part = jana_time_get_hours (time1);
	if (time1_part < hours)
		return -1;
	else if (time1_part > hours)
		return 1;
		
	/* Check minutes */
	time1_part = jana_time_get_minutes (time1);
	if (time1_part < minutes)
		return -1;
	else if (time1_part > minutes)
		return 1;
	
	/* Check seconds */
	time1_part = jana_time_get_seconds (time1);
	if (time1_part < seconds)
		return -1;
	else if (time1_part > seconds)
		return 1;
	
	return 0;
}

JanaTime *
jana_utils_time_copy (JanaTime *source, JanaTime *dest)
{
	gchar *tzname;
	
	tzname = jana_time_get_tzname (source);
	jana_time_set_tzname (dest, tzname);
	g_free (tzname);
	
	jana_time_set_offset (dest, jana_time_get_offset (source));
	
	jana_time_set_year (dest, jana_time_get_year (source));
	jana_time_set_month (dest, jana_time_get_month (source));
	jana_time_set_day (dest, jana_time_get_day (source));

	jana_time_set_hours (dest, jana_time_get_hours (source));
	jana_time_set_minutes (dest, jana_time_get_minutes (source));
	jana_time_set_seconds (dest, jana_time_get_seconds (source));

	jana_time_set_isdate (dest, jana_time_get_isdate (source));
	
	return dest;
}

void
jana_utils_time_diff (JanaTime *t1, JanaTime *t2, gint *year, gint *month,
		      gint *day, gint *hours, gint *minutes, gint *seconds)
{
	/* FIXME: Accumulate time in nearest measurement when pointers are
	 *        unset.
	 */
	if (year) *year = jana_time_get_year (t2) - jana_time_get_year (t1);
	if (month) *month = jana_time_get_month (t2) - jana_time_get_month (t1);
	if (day) *day = jana_time_get_day (t2) - jana_time_get_day (t1);
	if (hours) *hours = jana_time_get_hours (t2) - jana_time_get_hours (t1);
	if (minutes) *minutes = jana_time_get_minutes (t2) -
		jana_time_get_minutes (t1);
	if (seconds) *seconds = jana_time_get_seconds (t2) -
		jana_time_get_seconds (t1);
}

void
jana_utils_time_adjust (JanaTime *time, gint year, gint month, gint day,
			gint hours, gint minutes, gint seconds)
{
	if (seconds) jana_time_set_seconds (
		time, jana_time_get_seconds (time) + seconds);
	if (minutes) jana_time_set_minutes (
		time, jana_time_get_minutes (time) + minutes);
	if (hours) jana_time_set_hours (
		time, jana_time_get_hours (time) + hours);
	if (day) jana_time_set_day (
		time, jana_time_get_day (time) + day);
	if (month) jana_time_set_month (
		time, jana_time_get_month (time) + month);
	if (year) jana_time_set_year (
		time, jana_time_get_year (time) + year);
}

gboolean
jana_utils_duration_contains (JanaDuration *duration, JanaTime *time)
{
	if ((jana_utils_time_compare (time, duration->start, FALSE) >= 0) &&
	    (jana_utils_time_compare (time, duration->end, FALSE) < 0))
		return TRUE;
	else
		return FALSE;
}

JanaEvent *
jana_utils_event_copy (JanaEvent *source, JanaEvent *dest)
{
	gchar *string, **string_list;
	JanaTime *time;

	string = jana_event_get_summary (source);
	jana_event_set_summary (dest, string);
	g_free (string);
	
	string = jana_event_get_description (source);
	jana_event_set_description (dest, string);
	g_free (string);
	
	string = jana_event_get_location (source);
	jana_event_set_location (dest, string);
	g_free (string);
	
	time = jana_event_get_start (source);
	jana_event_set_start (dest, time);
	g_object_unref (time);

	time = jana_event_get_end (source);
	jana_event_set_end (dest, time);
	g_object_unref (time);
	
	string_list = jana_event_get_categories (source);
	jana_event_set_categories (dest, (const gchar **)string_list);
	g_strfreev (string_list);
	
	if (jana_event_supports_alarm (dest) &&
	    jana_event_supports_alarm (source)) {
		if (jana_event_has_alarm (source)) {
			time = jana_event_get_alarm_time (source);
			jana_event_set_alarm (dest, time);
			g_object_unref (time);
		} else
			jana_event_set_alarm (dest, NULL);
	}

	if (jana_event_supports_recurrence (dest) &&
	    jana_event_supports_recurrence (source)) {
		if (jana_event_has_recurrence (source)) {
			JanaRecurrence *recur =
				jana_event_get_recurrence (source);
			jana_event_set_recurrence (dest, recur);
			jana_recurrence_free (recur);
		} else
			jana_event_set_recurrence (dest, NULL);
	}

	if (jana_event_supports_exceptions (dest) &&
	    jana_event_supports_exceptions (source)) {
		if (jana_event_has_exceptions (source)) {
			GList *exceptions = jana_event_get_exceptions (source);
			jana_event_set_exceptions (dest, exceptions);
			jana_exceptions_free (exceptions);
		} else
			jana_event_set_exceptions (dest, NULL);
	}
	
	return dest;
}

GList *
jana_utils_event_get_instances (JanaEvent *event, JanaTime *range_start,
				JanaTime *range_end)
{
	JanaTime *instance_start, *instance_end, *end;
	GList *instances = NULL;
	
	/* TODO: Recurrence support, exception support */
	
	/* Split instances into days */
	instance_start = jana_event_get_start (event);
	instance_end = jana_time_duplicate (instance_start);
	jana_time_set_isdate (instance_end, TRUE);
	jana_time_set_day (instance_end, jana_time_get_day (instance_end) + 1);
	
	end = jana_event_get_end (event);

	while (jana_utils_time_compare (instance_start, end, FALSE) < 0) {
		JanaDuration *instance;
		if (jana_utils_time_compare (instance_end, end, FALSE) > 0) {
			g_object_unref (instance_end);
			instance_end = jana_time_duplicate (end);
		}
		
		if (((!range_end) || (jana_utils_time_compare (
		     instance_start, range_end, FALSE) < 0)) &&
		    ((!range_start) || (jana_utils_time_compare (
		     instance_end, range_start, FALSE) >= 0))) {
			instance = jana_duration_new (
				instance_start, instance_end);
			instances = g_list_append (instances, instance);
		}
		
		jana_time_set_day (instance_start, jana_time_get_day (
			instance_start) + 1);
		jana_time_set_day (instance_end, jana_time_get_day (
			instance_end) + 1);
		jana_time_set_isdate (instance_start, TRUE);
	}
	
	g_object_unref (instance_start);
	g_object_unref (instance_end);
	g_object_unref (end);
	
	return instances;
}

void
jana_utils_instance_list_free (GList *instances)
{
	while (instances) {
		jana_duration_free ((JanaDuration *)instances->data);
		instances = g_list_delete_link (instances, instances);
	}
}

gchar *
jana_utils_get_local_tzname ()
{
	static gboolean tzset_called = FALSE;
	
	if (!tzset_called) {
		tzset ();
	}

	return g_strdup_printf ("%s%s%s", tzname[0],
		tzname[1] ? "/" : "",
		tzname[1] ? tzname[1] : "");
}
